/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.impl.work;

import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.MBeanServer;
import javax.management.ObjectName;

import com.zfabrik.impl.work.jmx.ThreadPoolStatistics;
import com.zfabrik.impl.work.jmx.ThreadPoolStatisticsMBean;
import com.zfabrik.util.function.ThrowingRunnable;
import com.zfabrik.util.function.ThrowingSupplier;
import com.zfabrik.work.IThreadPool;
import com.zfabrik.work.WorkManager;
import com.zfabrik.work.WorkUnit;

/**
 * A thread pool implementation that has the following qualities:
 * <ul>
 * <li>All work is performed in a {@link WorkUnit}</li>
 * <li>Work from other threads may be performed under the concurrency limits of this threadpool</li>
 * <li>The pool is registered with JMX</li>
 * <ul>
 */
public class ThreadPoolImpl implements IThreadPool {
	private final Logger log, perfLog;
	
	private Object lock = new Object();
	// the actual work queue
	public LinkedBlockingQueue<Runnable> queue = new LinkedBlockingQueue<>();
	// We track work started from this pool
	private ThreadLocal<Boolean> ourThread = new ThreadLocal<>(); 
	// we use a regular jvm thread pool
	private ThreadPoolExecutor executor = new ThreadPoolExecutor(1, Integer.MAX_VALUE, 60, TimeUnit.SECONDS, queue);
	// an id as provided during registration
	private String id;
		
	/**
	 * This is the specified maximal concurrency. It is observed
	 * when processing asynchronous work. When blocking work from a thread of this pool, 
	 * we grant some excess to avoid potential thread starvation.  
	 */
	public int maxConcurrencyAllowed; // hard limit: max concurrency
	/**
	 * Max concurrency reached since last reset
	 */
	public int maxConcurrencyReached = 0;
	/**
	 * Current concurrency
	 */
	public int concurrency; // current concurrency
	/**
	 * Extra granted excess count of threads 
	 * to make sure a blocking execution may not lead to starvation.
	 * This is also used to account for external threads and may turn negative
	 */
	private int excess;
	/**
	 * How many execute's are waiting on capacity
	 */
	private int waiters;
	
	/**
	 * Monitoring
	 */
	private ObjectName objectName;
	private ThreadPoolStatisticsMBean mbean;
	public Date lastResetTime = null;
	/**
	 * Max length of queue attained since last reset
	 */
	public int maxTaskQueueLength = 0;
	public long tasksCompleted = 0;
	public long lastDispatch = 0;

	/**
	 * Create threadpool implementation and register with JMX
	 */
	public ThreadPoolImpl(String id) {
		this.id = id;
		this.maxConcurrencyAllowed = 1;
		this.log = Logger.getLogger("system.work.threadpool." + id);
		this.perfLog = Logger.getLogger("system.performance.system.work.threadpool." + id);
		try {
			MBeanServer platformMBeanServer = ManagementFactory.getPlatformMBeanServer();
			this.objectName = ObjectName.getInstance("zfabrik:type="+ThreadPoolImpl.class.getName()+",name=" + this.id);
			this.mbean=new ThreadPoolStatistics(this.lock, this);
			platformMBeanServer.registerMBean(this.mbean, this.objectName);
		} catch (Exception e) {
			throw new IllegalStateException("mbean registration for thread pool failed", e);
		}
	}
	
	public ObjectName getObjectName() {
		return objectName;
	}
	
	public ThreadPoolStatisticsMBean getMbean() {
		return mbean;
	}

	@Override
	public void close() {
		synchronized (lock) {
			try {
				this.executor.shutdownNow();
			} catch (Exception e) {
				log.log(Level.WARNING,"Error during executor shutdown",e);
			} finally {
				try {
					ManagementFactory.getPlatformMBeanServer().unregisterMBean(objectName);
				} catch (Exception e) {
					log.log(Level.WARNING,"Error during MBean unregistration",e);
				} finally {
					Optional.ofNullable(WorkManager.query()).ifPresent(wm->wm.releaseThreadPool(this.id));
				}
			}
		}
	}
	
	public boolean isClosed() {
		synchronized (lock) {
			return this.executor.isShutdown();
		}
	}

	// the actual work handling
	private List<Future<?>> work(boolean block, Collection<? extends Runnable> runnables) {
		if (runnables.isEmpty()) {
			return List.of();
		}
		List<Future<?>> result = new ArrayList<>(runnables.size());
		log.finer(()->"Putting collection of " + runnables.size() + " into execution");
		// is it us?
		boolean pt = isPoolThread(Thread.currentThread());
		try {
			if (block && pt) {
				synchronized (lock) {
					// #2040: grant some excess so we may not starve
					this._incExcess();
				}
			}
			// submit
			synchronized (lock) {
				runnables.forEach(r->{
					if (r instanceof ThreadExecutionGate) {
						// no work unit or concurrency accounting required,
						// as the actual work does it already. Just go for it
						executor.submit((ThreadExecutionGate)r);
					} else {
						result.add(executor.submit(()->runInWorkUnit(r)));
					}
				});
				int s = this.queue.size();
				if (s>this.maxTaskQueueLength) {
					this.maxTaskQueueLength = s;
				}
			}
			if (block) {
				// wait for futures
				for (Future<?> f : result) {
					try {
						f.get();
					} catch (Exception e) {
						throw new RuntimeException(e);
					}
				}
			}
		} finally {
			if (block && pt) {
				synchronized (lock) {
					this._decExcess();
				}
			}
		}
		return result;
	}


	// inc concurrency and optionally block, if exceeding limits
	private void _incConcurrency(boolean blockAtLimit) {
		if (blockAtLimit) {
			// as the thread pool executor is not really honoring core pool size = 0,
			// we put a max concurrency limit here by waiting in case of excess
			while (this.maxConcurrencyAllowed+this.excess<=this.concurrency) {
				try {
					this.waiters++;
					this.lock.wait();
					this.waiters--;
					if (this.waiters>0) {
						// others might try as well
						this.lock.notify();
					}
				} catch (InterruptedException e) {
					throw new IllegalStateException("Wait for capacity interrupted");
				}
			}
		}
		this.concurrency++;
		if (this.concurrency>this.maxConcurrencyReached) {
			this.maxConcurrencyReached=this.concurrency;
		}
	}
	
	// dec. concurrency and notify waiting parties
	private void _decConcurrency() {
		concurrency--;
		// trigger waiters
		if (this.waiters>0){
			this.lock.notify();
		}
	}

	
	// inc excess and so max pool size
	private void _incExcess() {
		this.excess++;
		_applyMax();
	}

	// dec excess and so max pool size
	private void _decExcess() {
		this.excess--;
		_applyMax();
	}

	private void _applyMax() {
		this.log.fine(()->"Setting pool size to "+this.maxConcurrencyAllowed+" plus excess "+this.excess);
		// NOTE: we work with the core pool size. The max pool size is 
		// only effective, of the queue is full (never happening in our case)
		int m = Math.max(this.maxConcurrencyAllowed+this.excess,0);
		// NOTE: We might set it to zero, but that is not really effective,
		// and one thread may still always run. That is 
		// why we have the extra waiting mechanism in #_incConcurrency
		this.executor.setCorePoolSize(m);
		// some might be waiting
		if (this.waiters>0) {
			this.lock.notify();
		}
	}
	
	/**
	 * run a supplier with {@link WorkUnit}. Everything goes through here!
	 * Optionally have it block at limits.
	 * 
	 * See {@link #getAs(ThrowingSupplier, boolean)} for use cases
	 */
	private <T, E extends Exception> T getInWorkUnit(ThrowingSupplier<T,E> c, boolean blockAtLimit) throws E {
		long start = 0;
		try {
			synchronized (lock) {
				this._incConcurrency(blockAtLimit);
			}
			this.ourThread.set(true);
			start = System.currentTimeMillis();
			log.finer(()->"executing " + c);
			return WorkUnit.supply(c);
		} finally {
			this.ourThread.set(false);
			synchronized (lock) {
				_decConcurrency();
				this.tasksCompleted++;
			}
			if (perfLog.isLoggable(Level.FINE)) {
				long durationTotal = System.currentTimeMillis() - start;
				perfLog.fine("system.work.threadpool." + id + ",total work unit," + c + ",duration," + durationTotal + ",ms");
				perfLog.fine("system.work.threadpool." + id + ",total runnable," + c + ",duration," + durationTotal + ",ms");
			}
		}
	}

	/**
	 * run a runnable with workunit accounting for concurrency and observing limits
	 */
	private void runInWorkUnit(Runnable c) {
		getInWorkUnit(()->{
				c.run();
				return null;
			},
			true
		);
	}
	
	public int getExcess() {
		synchronized (lock) {
			return excess;
		}
	}
	
	//
	// interface
	//

	@Override
	public List<Future<?>> execute(boolean block, Collection<? extends Runnable> runnables) {
		return this.work(block, runnables);
		
	}

	@Override
	public List<Future<?>> execute(boolean block, Runnable... runnables) {
		return this.work(block, List.of(runnables));
	}

	@Override
	public void setMaxConcurrency(int maxconcurrency) {
		synchronized (lock) {
			this.maxConcurrencyAllowed = maxconcurrency;
			this._applyMax();
		}
	}

	@Override
	public int getMaxConcurrency() {
		synchronized (lock) {
			return this.maxConcurrencyAllowed;
		}
	}

	@Override
	public int getMaxAchievedConcurrency() {
		synchronized (lock) {
			return this.maxConcurrencyReached;
		}
	}

	// 
	
	/**
	 * The thread is from the pool or associated via {@link #runAs(ThrowingRunnable, boolean)} and the like
	 * @param t
	 * @return
	 */
	@Override
	public boolean isPoolThread(Thread t) {
		return this.ourThread.get()==Boolean.TRUE;
	}

	@Override
	public void executeAs(Runnable runnable, boolean observeMaxConcurrency) {
		runAs(
			runnable::run, 
			observeMaxConcurrency
		);
	}

	@Override
	public <T> T executeAs(Callable<T> callable, boolean observeMaxConcurrency) throws Exception {
		return getAs(
			callable::call, 
			observeMaxConcurrency
		);
	}
	
	/**
	 * Run a {@link ThrowingSupplier} as accounted and with {@link WorkUnit} of this
	 * thread pool.
	 */
	@Override
	public <T, E extends Exception> T getAs(ThrowingSupplier<T,E> supplier, boolean observeMaxConcurrency) throws E {
		if (isPoolThread(Thread.currentThread()) || !observeMaxConcurrency) {
			// just go there - we don't count recursive association as increase of concurrency 
			return WorkUnit.supply(supplier);
		} else	{
			// we want to have it accounted as concurrency in this pool,
			// but it is not from this pool, so, if the current
			// max concurrency is already exceeded, we need to wait.
			// To model that we put a gate into the pool that opens
			// once the pool runs it.
			// As this requires thread synchronization, this may be quiet expensive
			// and we attempt an optimization first
			ThreadExecutionGate qg;
			synchronized (lock) {
				if (this.queue.size()+this.executor.getActiveCount()<this.maxConcurrencyAllowed+this.excess) {
					// we have some air to breath, run it directly
					// but count it as negative excess
					this._decExcess();
					qg=null;
					log.finer(()->"Running As without gate mediation");
				} else {
					// need to have it mediated
					log.finer(()->"Running As with gate mediation");
					qg=new ThreadExecutionGate();
				}
			}
			if (qg!=null) {
				// start mediation by a gate
				this.execute(false, List.of(qg));
				// and pass execution into the gate
				return qg.work(()->getInWorkUnit(supplier, true));
			} else {
				try {
					// run directly
					return getInWorkUnit(supplier, false);
				} finally {
					synchronized (lock) {
						this._incExcess();
					}					
				}
			}
		}
	}
	
	@Override
	public <E extends Exception> void runAs(ThrowingRunnable<E> runnable, boolean maxconcurrency) throws E {
		this.getAs(()->{
				runnable.run();
				return null;
			}, 
			maxconcurrency
		);
	}
	
}