/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */

package com.zfabrik.dev.z2jupiter.internal.transfer;

import java.io.Serializable;
import java.util.function.Function;

import org.junit.platform.engine.TestDescriptor;
import org.junit.platform.engine.UniqueId;
import org.junit.platform.launcher.TestExecutionListener;
import org.junit.platform.launcher.TestIdentifier;

import com.zfabrik.dev.z2jupiter.internal.client.Z2JupiterClientTestDescriptor;
import com.zfabrik.dev.z2jupiter.internal.util.UniqueIds;

/**
 * Payload of event transmission back to client and handling on client
 */
public class Z2JupiterListenerEventDto implements Serializable {
	private static final long serialVersionUID = -5828485812907058733L;

	/**
	 * Event type and dispatch to client side {@link TestExecutionListener}
	 */
	public static enum Type {
		testPlanExecutionStarted, testPlanExecutionFinished, dynamicTestRegistered, executionSkipped, executionStarted,
		executionFinished, reportingEntryPublished;
	}

	private boolean last;
	private Type type;
	private Z2JupiterTestResultDto testResult;
	private Z2JupiterTestIdentifierDto testIdentifier;
	private String reason;
	private Z2JupiterReportEntryDto reportEntry;
	
	private transient Z2JupiterClientTestDescriptor descriptor;
	private transient TestIdentifier identifier;
	// must be set for conversions on client!
	private transient Function<UniqueId,TestDescriptor> resolver; 
	
	public Z2JupiterListenerEventDto() {}
	
	/**
	 * Create stop event dto (sent as end of events)
	 */
	public static Z2JupiterListenerEventDto last() {
		Z2JupiterListenerEventDto that = new Z2JupiterListenerEventDto();
		that.last = true;
		return that;
	}

	/**
	 * Must be set on client, when processing to test identifier or testdescriptors
	 * @param clientTestPlan
	 * @return
	 */
	public Z2JupiterListenerEventDto withParentDescriptorResolver(Function<UniqueId,TestDescriptor> resolver) {
		this.resolver = resolver;
		return this;
	}
	
	/**
	 * Create JUnit5 descriptor given the event data
	 */
	public Z2JupiterClientTestDescriptor toJUnit5Descriptor() {
		return toJUnit5Descriptor(null);
	}
	
	/**
	 * Create JUnit5 descriptor given the event data and apply a prefix to the uniqueId
	 */
	public Z2JupiterClientTestDescriptor toJUnit5Descriptor(UniqueId prefix) {
		if (descriptor==null) {
			if (this.resolver==null) {
				throw new IllegalStateException("No resolver set!");
			}
			this.descriptor = new Z2JupiterClientTestDescriptor(this.testIdentifier, prefix);
			if (this.testIdentifier.getParentId()!=null) {
				// fix parent
				this.descriptor.setParent(
					this.resolver.apply(
						UniqueIds.concat(prefix, UniqueId.parse(this.testIdentifier.getParentId()))
					)
				);
			}
		}
		return this.descriptor;
	}

	/**
	 * Create JUnit5 identifier given the event data
	 */
	public TestIdentifier toJUnit5TestIdentifier() {
		if (identifier==null) {
			this.identifier = TestIdentifier.from(toJUnit5Descriptor());
		}
		return this.identifier;
	}
	
	public Type getType() {
		return type;
	}
	
	public void setType(Type type) {
		this.type = type;
	}
	
	public Z2JupiterListenerEventDto withType(Type type) {
		setType(type);
		return this;
	}
	
	public Z2JupiterTestResultDto getTestResult() {
		return testResult;
	}

	public void setTestResult(Z2JupiterTestResultDto testResult) {
		this.testResult = testResult;
	}

	public Z2JupiterListenerEventDto withTestResult(Z2JupiterTestResultDto testResult) {
		setTestResult(testResult);
		return this;
	}

	public Z2JupiterTestIdentifierDto getTestIdentifier() {
		return testIdentifier;
	}
	
	public void setTestIdentifier(Z2JupiterTestIdentifierDto testIdentifier) {
		this.testIdentifier = testIdentifier;
	}
	
	public Z2JupiterListenerEventDto withTestIdentifier(Z2JupiterTestIdentifierDto testIdentifier) {
		setTestIdentifier(testIdentifier);
		return this;
	}

	public String getReason() {
		return reason;
	}

	public void setReason(String reason) {
		this.reason = reason;
	}
	
	public Z2JupiterListenerEventDto withReason(String reason) {
		setReason(reason);
		return this;
	}
	
	public Z2JupiterReportEntryDto getReportEntry() {
		return reportEntry;
	}
	
	public void setReportEntry(Z2JupiterReportEntryDto reportEntry) {
		this.reportEntry = reportEntry;
	}

	public Z2JupiterListenerEventDto withReportEntry(Z2JupiterReportEntryDto reportEntry) {
		this.setReportEntry(reportEntry);
		return this;
	}
	
	public boolean isLast() {
		return last;
	}

	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (last ? 1231 : 1237);
		result = prime * result + ((reason == null) ? 0 : reason.hashCode());
		result = prime * result + ((reportEntry == null) ? 0 : reportEntry.hashCode());
		result = prime * result + ((testIdentifier == null) ? 0 : testIdentifier.hashCode());
		result = prime * result + ((testResult == null) ? 0 : testResult.hashCode());
		result = prime * result + ((type == null) ? 0 : type.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Z2JupiterListenerEventDto other = (Z2JupiterListenerEventDto) obj;
		if (last != other.last)
			return false;
		if (reason == null) {
			if (other.reason != null)
				return false;
		} else if (!reason.equals(other.reason))
			return false;
		if (reportEntry == null) {
			if (other.reportEntry != null)
				return false;
		} else if (!reportEntry.equals(other.reportEntry))
			return false;
		if (testIdentifier == null) {
			if (other.testIdentifier != null)
				return false;
		} else if (!testIdentifier.equals(other.testIdentifier))
			return false;
		if (testResult == null) {
			if (other.testResult != null)
				return false;
		} else if (!testResult.equals(other.testResult))
			return false;
		if (type != other.type)
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "Z2JupiterListenerEventDto [last=" + last + ", type=" + type + ", testResult=" + testResult + ", testIdentifier=" + testIdentifier
				+ ", reason=" + reason + ", reportEntry=" + reportEntry + "]";
	}
	
}
