/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.dev.z2jupiter.impl;

import static com.zfabrik.dev.z2jupiter.internal.engine.Z2JupiterTestEngine.withoutTestEngine;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.logging.Logger;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.StringUtils;
import org.junit.platform.launcher.Launcher;
import org.junit.platform.launcher.TestPlan;
import org.junit.platform.launcher.core.LauncherFactory;

import com.zfabrik.components.java.JavaComponentUtil;
import com.zfabrik.dev.z2jupiter.internal.client.Z2JupiterClient;
import com.zfabrik.dev.z2jupiter.internal.transfer.Z2JupiterListenerEventDto;
import com.zfabrik.util.threading.ThreadUtil;

/**
 * Generic abstract request handling for Z2Jupiter
 */
public abstract class Z2JupiterRequest {
	protected static final Logger LOG = Logger.getLogger(Z2JupiterRequest.class.getName());

	protected HttpServletRequest request; 
	protected HttpServletResponse response;
	protected String componentName;
	protected ClassLoader loader;
	
	protected Z2JupiterRequest(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		super();
		this.request = request;
		this.response = response;
		this.requiresComponentName();
		this.loader = JavaComponentUtil.getPrivateLoaderOfJavaComponent(this.componentName, null);
		if (this.loader==null) {
			throw new IllegalArgumentException("Component "+this.componentName+" does not resolve to code");
		}
	}

	private void requiresComponentName() {
		this.componentName = StringUtils.trimToNull(request.getParameter(Z2JupiterClient.PARAM_COMPONENT_NAME));
		if (this.componentName==null) {
			throw new IllegalArgumentException("Missing parameter componentName");
		}
	}
	
	/**
	 * Run in context of test module with disabled Z2U
	 */
	protected void runInServerContext(Runnable run) {
		ThreadUtil.cleanContextRun(
			this.loader,
			()->withoutTestEngine(run)
		);
	}
	
	/**
	 * Retrieve POST data 
	 */
	protected <T> T getPOST(Class<T> clz) {
		requirePOST();
		String contentType = request.getContentType();
		if (!Z2JupiterClient.APPLICATION_OCTET_STREAM.equals(StringUtils.trimToEmpty(contentType.replace(" ", "").toLowerCase()))) {
			throw new IllegalArgumentException("Unsupported content type "+contentType);
		}
		return receive(request ,clz);
	}

	/**
	 * Get JUnit launcher
	 */
	protected Launcher launcher() {
		return LauncherFactory.create();
	}

	/**
	 * Json to output in response
	 */
	protected void send(Object o) {
		try (ObjectOutputStream out = new ObjectOutputStream(response.getOutputStream())){
			out.writeObject(o);
			response.getOutputStream().flush();
		} catch (Exception e) {
			throw new IllegalArgumentException("Marshalling error "+e.getMessage(),e);
		}
	}

	/**
	 * Json from input in request
	 */
	protected <T> T receive(HttpServletRequest request,  Class<T> clz) {
		try (ObjectInputStream in = new ObjectInputStream(request.getInputStream())) {
			return clz.cast(in.readObject());
		} catch (ClassNotFoundException | IOException e) {
			throw new IllegalArgumentException("Unmarshalling error "+e.getMessage(),e);
		}
	}
	
	/**
	 * Prepare response for output
	 */
	protected HttpServletResponse prepareResponse() {
		response.setContentType(Z2JupiterClient.APPLICATION_OCTET_STREAM);
		response.setBufferSize(0);
		return response;
	}
	
	/**
	 * Require HTTP method
	 */
	protected void requirePOST() {
		if (!Z2JupiterClient.HTTP_METHOD_POST.equals(request.getMethod())) {
			throw new IllegalArgumentException("Requiring POST for "+request.getRequestURI());
		}
	}

	/**
	 * Require HTTP method
	 */
	protected void requireGET() {
		if (!Z2JupiterClient.HTTP_METHOD_GET.equals(request.getMethod())) {
			throw new IllegalArgumentException("Requiring GET for "+request.getRequestURI());
		}
	}

	/**
	 * Run a given testplan, with disabled Z2JupiterEngine, and stream events back to response
	 * in JSON array
	 */
	protected void runTestPlanAndStreamEventsToResponse(TestPlan tp) {
		prepareResponse();
		try (ObjectOutputStream out = new ObjectOutputStream(response.getOutputStream())) {
			/*
			 * Execute with local launcher and send
			 * events back
			 */
			this.launcher().execute(
				tp, 
				// listener reporting back to 
				// the z2 jupiter client
				new Z2JupiterListenerAdapter(e->{
					try {
						out.writeObject(e);
					} catch (IOException ioe) {
						throw new IllegalArgumentException("Marshalling error "+ioe.getMessage(),ioe);
					}
				}),
				// listener enforcing dependencies
				new Z2DependencyPreparationListener()
			);
			// send a termination event
			out.writeObject(Z2JupiterListenerEventDto.last());
		} catch (IOException ioe) {
			throw new IllegalArgumentException("Marshalling error "+ioe.getMessage(),ioe);
		}
	}

}