/*
 * z2-Environment
 * 
 * Copyright(c) ZFabrik Software GmbH & Co. KG
 * 
 * contact@zfabrik.de
 * 
 * http://www.z2-environment.eu
 */
package com.zfabrik.tx;

import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.transaction.NotSupportedException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.UserTransaction;

import com.zfabrik.util.function.ThrowingRunnable;
import com.zfabrik.util.function.ThrowingSupplier;
import com.zfabrik.util.threading.ThreadUtil;
import com.zfabrik.work.WorkUnit;

/**
 * This class provides utilities around transaction management with Z2's built-in JTA implementation for
 * easier transaction demarcation.
 * 
 * In particular, using {@link #get(UserTransaction, ThrowingSupplier)}, {@link #run(UserTransaction, Callable)}, or {@link #run(UserTransaction, ThrowingRunnable)}
 * it is not necessary to using complicated try-catch constructions making sure a transaction is orderly completed.
 * <p>
 * Also, these methods support nesting of invocations. 
 * <p>
 * In that case, only the outer-most invocation will complete the transaction. 
 * <p>
 * Leaving any invocation (nested or not) will set the current transaction to rollback only. That is also the preferred
 * way of forcing a rollback of a transaction.
 * <p>
 * Invoking an explicit commit or rollback on the user transaction while in a call to any of the utility methods 
 * above will normally unassociate a given transaction from the thread and lead to exceptions when leaving the {@link TransactionUtil} 
 * methods. In other words: Do not use the {@link UserTransaction} other than checking on transaction status.
 */
public class TransactionUtil {
	
	/**
	 * Do not allow an outer active transaction. 
	 */
	public static void never(UserTransaction ut) {
		try {
			if (ut.getStatus()!=javax.transaction.Status.STATUS_NO_TRANSACTION) {
				throw new IllegalStateException("No ongoing transaction allowed!");
			}
		} catch (SystemException e) {
			throw new IllegalStateException(e);
		}
	}

	/**
	 * Require an outer active transaction. 
	 */
	public static void mandatory(UserTransaction ut) {
		try {
			if (ut.getStatus()!=javax.transaction.Status.STATUS_ACTIVE) {
				throw new IllegalStateException("An ongoing transaction is required!");
			}
		} catch (SystemException e) {
			throw new IllegalStateException(e);
		}
	}
	
	/**
	 * The TransactionUtilException is used to wrap system style exceptions possibly occurring when
	 * invoking transaction management methods, but normally not of interest to application, 
	 * when invoking {@link TransactionUtil#get(UserTransaction, ThrowingSupplier)},
	 * {@link TransactionUtil#run(UserTransaction, Callable)}, {@link TransactionUtil#run(UserTransaction, ThrowingRunnable)}.
	 * 
	 * In that case this helps distinguishing application created {@link Exception}s from transaction management
	 * related {@link Exception}s.
	 */
	public static class TransactionUtilException extends RuntimeException {
		private static final long serialVersionUID = 1L;

		public TransactionUtilException(Throwable cause) {
			super("System error during transaction util execution",cause);
		}
	}
	
	/**
	 * Use this method to execute work in a transaction. This is just a utility that helps implementing
	 * for example Transaction Servlet Filters that begin and end transactions at the request boundaries.
	 * <p>
	 * If there is an ongoing transaction already, this method will not begin a new one. Any exception caught in that case will
	 * set that new transaction to rollback. 
	 * <p>
	 * See also {@link ThreadUtil#confine(Callable, Class)} on how to handle checked exception cases
	 * gracefully.
	 * 
	 * @param ut the {@link UserTransaction} instance of the transaction management system in place
	 * @param callable The {@link Callable} implementation that defines the work.
	 * @return The result of the callable invocation.
	 * @throws Exception
	 */
	public static <V> V run(UserTransaction ut, Callable<V> callable) throws Exception {
		return get(ut,callable::call);
	}

	/**
	 * Use this method to execute work in a transaction that returns a result. 
	 * <p>
	 * If there is an ongoing transaction already, this method will not begin a new one. Any exception caught in that case will
	 * set that new transaction to rollback. 
	 * <p>
	 * This method relies on {@link ThrowingSupplier} to give the caller control over what exceptions may be propagated.
	 * <p>
	 * Any other checked exceptions that may be thrown by underlying transaction management methods, such as {@link SystemException}, {@link NotSupportedException}
	 * will be wrapped into a {@link TransactionUtilException}.
	 * 
	 * @param ut the {@link UserTransaction} instance of the transaction management system in place
	 * @param supplier The {@link ThrowingSupplier} implementation that defines the work.
	 * @return The result of the supplier invocation.
	 * @throws Exception
	 */
	public static <V, E extends Exception> V get(UserTransaction ut, ThrowingSupplier<V, E> supplier) throws E, TransactionUtilException {
		// Note: We control the workunit completely and do
		// not use WorkUnit::supply as that would set the 
		// outer work unit to rollback only when leaving with
		// an exception (#2180)
		WorkUnit current = WorkUnit.initCurrent();
		try {
			int status;
			try {
				status = ut.getStatus();
			} catch (Exception syse) {
				throw new TransactionUtilException(syse);
			}
			if (status!=Status.STATUS_ACTIVE) {
				try {
					ut.begin();
				} catch (Exception be) {
					throw new TransactionUtilException(be);
				}
				boolean success=false;
				try {
					V v = supplier.get();
					success=true;
					return v;
				} finally {
					try {
						if (!success) {
							logger.log(Level.WARNING,"Error at transaction boundary. Setting transaction to rollbackonly");
							ut.setRollbackOnly();
						}
						if (ut.getStatus()==Status.STATUS_ACTIVE) {
							ut.commit();
						} else {
							ut.rollback();
						}
					} catch (Exception syse) {
						throw new TransactionUtilException(syse);
					}
				}
			} else {
				// there is a tx ongoing already. Simply work.
				return supplier.get();
			}
		} finally {
			current.close();
		}
	}
	
	/**
	 * Use this method to execute work in a transaction that returns no result. 
	 * <p>
	 * If there is an ongoing transaction already, this method will not begin a new one. Any exception caught in that case will
	 * set that new transaction to rollback. 
	 * <p>
	 * This method relies on {@link ThrowingSupplier} to give the caller control over what exceptions may be propagated.
	 * <p>
	 * Any other checked exceptions that may be thrown by underlying transaction management methods, such as {@link SystemException}, {@link NotSupportedException}
	 * will be wrapped into a {@link TransactionUtilException}.
	 * 
	 * @param ut the {@link UserTransaction} instance of the transaction management system in place
	 * @param runnable The {@link ThrowingSupplier} implementation that defines the work.
	 * @return The result of the supplier invocation.
	 * @throws Exception
	 */
	public static <E extends Exception> void run(UserTransaction ut, ThrowingRunnable<E> runnable) throws E, TransactionUtilException {
		get(
			ut,
			()->{
				runnable.run();
				return null;
			}
		);
	}

	
	private final static Logger logger = Logger.getLogger(TransactionUtil.class.getName());
}
